package cofh.cofhworld.init;

import cofh.cofhworld.CoFHWorld;
import cofh.cofhworld.util.ChunkCoord;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.THashSet;
import net.minecraft.nbt.NBTTagList;
import net.minecraft.world.World;
import net.minecraftforge.fml.common.eventhandler.SubscribeEvent;
import net.minecraftforge.fml.common.gameevent.TickEvent.Phase;
import net.minecraftforge.fml.common.gameevent.TickEvent.WorldTickEvent;
import net.minecraftforge.fml.relauncher.Side;

import java.util.ArrayDeque;
import java.util.Random;

public class WorldTickHandler {

	public static WorldTickHandler instance = new WorldTickHandler();

	public static TIntObjectHashMap<ArrayDeque<RetroChunkCoord>> chunksToGen = new TIntObjectHashMap<>();
	public static TIntObjectHashMap<ArrayDeque<ChunkCoord>> chunksToPreGen = new TIntObjectHashMap<>();

	// FIXME: put adding to these behind a function so we can remove the tick handler when there's nothing to do
	// size of the maps indicates how many dimensions are needing to gen/pregen, and will be 0 when no work is required

	private static byte pregenC, retroC;

	@SubscribeEvent
	public void tickEnd(WorldTickEvent event) {

		if (event.side != Side.SERVER) {
			return;
		}
		World world = event.world;
		int dim = world.provider.getDimension();

		if (event.phase == Phase.END) {
			ArrayDeque<RetroChunkCoord> chunks = chunksToGen.get(dim);

			if (chunks != null && chunks.size() > 0) {
				RetroChunkCoord r = chunks.pollFirst();
				ChunkCoord c = r.coord;
				if (retroC++ == 0 || chunks.size() < 3) {
					CoFHWorld.log.info("RetroGening " + c.toString() + ".");
				} else {
					CoFHWorld.log.debug("RetroGening " + c.toString() + ".");
				}
				retroC &= 31;
				long worldSeed = world.getSeed();
				Random rand = new Random(worldSeed);
				long xSeed = rand.nextLong() >> 2 + 1L;
				long zSeed = rand.nextLong() >> 2 + 1L;
				rand.setSeed(xSeed * c.chunkX + zSeed * c.chunkZ ^ worldSeed);
				WorldHandler.instance.generateWorld(rand, r, world, false);
				chunksToGen.put(dim, chunks);
			} else if (chunks != null) {
				chunksToGen.remove(dim);
			}
		} else {
			ArrayDeque<ChunkCoord> chunks = chunksToPreGen.get(dim);

			if (chunks != null && chunks.size() > 0) {
				ChunkCoord c = chunks.pollFirst();
				if (pregenC++ == 0 || chunks.size() < 5) {
					CoFHWorld.log.info("PreGening " + c.toString() + ".");
				} else {
					CoFHWorld.log.debug("PreGening " + c.toString() + ".");
				}
				pregenC &= 31;
				world.getChunkFromChunkCoords(c.chunkX, c.chunkZ);
			} else if (chunks != null) {
				chunksToPreGen.remove(dim);
			}
		}
	}

	public static class RetroChunkCoord {

		private static final THashSet<String> emptySet = new THashSet<>(0);
		public final ChunkCoord coord;
		public final THashSet<String> generatedFeatures;

		public RetroChunkCoord(ChunkCoord pos, NBTTagList features) {

			coord = pos;
			if (features == null) {
				generatedFeatures = emptySet;
			} else {
				int i = 0, e = features.tagCount();
				generatedFeatures = new THashSet<>(e);
				for (; i < e; ++i) {
					generatedFeatures.add(features.getStringTagAt(i));
				}
			}
		}

		@Override
		public boolean equals(Object o) {

			if (o instanceof RetroChunkCoord) {
				return ((RetroChunkCoord) o).coord.equals(coord);
			} else if (o instanceof ChunkCoord) {
				return o.equals(coord);
			}
			return false;
		}

		@Override
		public int hashCode() {

			return coord.hashCode();
		}
	}

}
